<?xml version="1.0"?>
<!--

  xdr-xsd-converter.xslt 1.2
  
  XSLT Style sheet for transforming XML Data Reduced (XDR) schemas to
  compliant W3C XML Schema syntax (Oct 24 CR draft).  XSLT currently
  requires the MSXSL script extension and therefore requires the 
  Microsoft September 2000 MSXML Web Release or later, which can be 
  downloaded at http://msdn.microsoft.com/xml.
  
  1.0.1: (5 Oct 2000)
    Changed RTF queries to xalan:nodeset().
    Added exclude-result-prefixes to support validation.
    Changed output DTD to official version (XMLSchemas.dtd/part2.dtd)
  
  1.1: (10 Oct 2000)
    Major rewrite to update to Sept 22 XSD draft.
    Fixed a bunch of bugs discovered in the process.

  1.2: (25 Oct 2000)
    Validates against Oct 24 CR XSD draft.
    Minor cleanup of source (generate XML decl, PUBLIC id, better
      modularization, regluarization of named templates, etc.)
    Preservation of most comments in the source.
    Bug fixes
      - UUID pattern now allows lowercase
      - enums derived from xsd:NMTOKEN (had NMTOKENS in one place)
      - elements with enums and attributes fixed
      - elements with length constraints and attributes fixed
    
  Author: Jonathan Marsh <jmarsh@microsoft.com>
  Copyright 2000 Microsoft Corp.
  
-->

<xsl:stylesheet version="1.0"
                xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                xmlns:xdr = "urn:schemas-microsoft-com:xml-data"
                xmlns:dt = "urn:schemas-microsoft-com:datatypes"
                xmlns:xalan="http://xml.apache.org/xalan"
                exclude-result-prefixes="xdr dt xalan">

<xsl:output method="xml" omit-xml-declaration="no" indent="yes"/>

<xsl:template match="node()"/>
<xsl:template match="comment()">
  <xsl:copy-of select="."/>
</xsl:template>

<!-- main document conversion template -->

<xsl:template match="/">

<!-- Note that all comments generated by this stylesheet are preceded
     by "[XDR-XSD]" to distinguish them from comments passed through
     from the source XDR. -->
<xsl:comment>
  [XDR-XSD] This schema automatically updated from an IE5-compatible XDR schema to W3C
  XML Schema by xdr-xsd-converter.xslt 1.2, available from Microsoft.
  Contact jmarsh@microsoft.com.

  Note: if a local copy of XMLSchema.dtd and datatypes.dtd are not
  available, use the official location of "http://www.w3.org/2000/10/XMLSchema.dtd"
  for the system id.
</xsl:comment>

<!-- <xsl:text disable-output-escaping="yes">
<![CDATA[<!DOCTYPE xsd:schema PUBLIC "-//W3C//DTD XMLSCHEMA 200010//EN"
          "XMLSchema.dtd" [
<!ENTITY % p 'xsd:'>
<!ENTITY % s ':xsd'>
]>]]>
</xsl:text> -->

  <xsl:apply-templates select="@*|node()"/>

</xsl:template>

<xsl:template match="xdr:Schema">
<xsd:schema elementFormDefault="qualified">
  
  <!-- discover references to external elements -->
  <xsl:variable name="prefixes">
    <xsl:for-each select=".//xdr:attribute[contains(@type,':')] | .//xdr:element[contains(@type,':')]">
      <prefix><xsl:value-of select="substring-before(@type,':')"/></prefix>
    </xsl:for-each>
  </xsl:variable>
  
  <!-- sort the prefixes, in preparation for eliminating duplicates -->
  <xsl:variable name="sorted-prefixes">
    <xsl:for-each select="xalan:nodeset($prefixes)/prefix">
      <xsl:sort select="."/>
      <xsl:copy-of select="."/>
    </xsl:for-each>
  </xsl:variable>
  
  <!-- eliminate duplicaates -->
  <xsl:variable name="unique-prefixes">
    <xsl:for-each select="xalan:nodeset($sorted-prefixes)/prefix">
      <xsl:variable name="here" select="position()"/>
      <xsl:if test="$here=1 or ../prefix[position() &lt; $here][1] != .">
         <xsl:copy-of select="."/>
      </xsl:if>
    </xsl:for-each>
  </xsl:variable>
  
  <!-- make namespace declarations for any external schemas -->
  <xsl:variable name="root" select="/"/>
  <xsl:variable name="namespaces">
    <dummy>
      <xsl:for-each select="xalan:nodeset($unique-prefixes)/prefix">
        <xsl:attribute name="{.}:dummy" namespace="namespace-uri($root//namespace::*[name()=current()][1])"/>
      </xsl:for-each>
    </dummy>
  </xsl:variable>

  <xsl:copy-of select="xalan:nodeset($namespaces)/namespace::*"/>
  
  <!-- make import statements for any external schemas -->
  <xsl:for-each select="xalan:nodeset($unique-prefixes)/prefix">
    <xsl:variable name="ns" select="concat('xmlns:',current())"/>
    <xsl:variable name="ns-uri" select="$root//@*[name()=$ns]"/>
    <xsd:import namespace="{$ns-uri}" schemaLocation="{substring-after($ns-uri,'x-schema:')}"/>
  </xsl:for-each>
          
  <xsl:call-template name="annotations">
    <xsl:with-param name="documentation-extras">
      <xsl:if test="@name">Schema name: <xsl:value-of select="@name"/></xsl:if>
    </xsl:with-param>
  </xsl:call-template>
  
  <xsl:apply-templates select="node()">
    <xsl:sort select="not(self::xdr:description)"/>
  </xsl:apply-templates>
  
  <!-- Some XDR datatypes don't have direct equivalents in XSD.
       If any of these types are used, place a datatype derivation
       in the result. -->
  <xsl:comment> [XDR-XSD] XDR datatype derivations </xsl:comment>
  
  <xsl:if test="//@dt:type='fixed.14.4'">
  <xsd:simpleType name="fixed.14.4">
    <xsd:restriction base="xsd:decimal">
      <xsd:scale value="4"/>
      <xsd:minInclusive value="-922337203685477.5808"/>
      <xsd:maxInclusive value="922337203685477.5807"/>
    </xsd:restriction>
  </xsd:simpleType>
  </xsl:if>
  
  <xsl:if test="//@dt:type='bin.base64'">
  <xsd:simpleType name="bin.base64">
    <xsd:restriction base="xsd:binary">
      <xsd:encoding value="base64"/>
    </xsd:restriction>
    <!---data encoded in Base64 notation -->
  </xsd:simpleType>
  </xsl:if>
  
  <xsl:if test="//@dt:type='bin.hex'">
  <xsd:simpleType name="bin.hex">
    <xsd:restriction base="xsd:binary">
      <xsd:encoding value="hex"/>
    </xsd:restriction>
  </xsd:simpleType>
  </xsl:if>

  <xsl:if test="//@dt:type='boolean'">
  <xsd:simpleType name="ISOBoolean">
    <xsd:restriction base="xsd:integer">
      <xsd:enumeration value="0"/>  <!--False -->
      <xsd:enumeration value="1"/>  <!--True -->
    </xsd:restriction>
  </xsd:simpleType>
  </xsl:if>

  <xsl:if test="//@dt:type='number'">
  <xsd:simpleType name="number">
    <xsd:restriction base="xsd:string">
      <xsd:pattern value="[+-]?\d*(\.\d*)?"/>
    </xsd:restriction>
  </xsd:simpleType>
  </xsl:if>

  <xsl:if test="//@dt:type='char'">
  <xsd:simpleType name="char">
    <xsd:restriction base="xsd:string">
      <xsd:length value="1"/>
    </xsd:restriction>
  </xsd:simpleType>        
  </xsl:if>
   
  <xsl:if test="//@dt:type='r4'">
  <xsd:simpleType name="r4">
    <xsd:restriction base="xsd:float">
      <xsd:minInclusive value="-3.40282347E+38" />
      <xsd:maxInclusive value="3.40282347E+38" />
    </xsd:restriction>
  </xsd:simpleType>
  </xsl:if>

  <xsl:if test="//@dt:type='dateTime'">
  <xsd:simpleType name="dateTime">
    <xsd:restriction base="xsd:timeInstant">
      <xsd:pattern value="\d*-\d\d-\d\dT\d\d:\d\d(:\d\d(\.\d{{0,9}})?)?"/>
    </xsd:restriction>
    <!---CCYY-MM-DDTHH:MM:SS.fffffffff Note no time zone. -->
  </xsd:simpleType>
  </xsl:if>
  
  <xsl:if test="//@dt:type='time'">
  <xsd:simpleType name="time">
    <xsd:restriction base="xsd:timeInstant">
      <xsd:pattern value="\d\d:\d\d(:\d\d)?"/>
    </xsd:restriction>
    <!---HH:MM:SS Note no time zone. -->
  </xsd:simpleType>
  </xsl:if>

  <xsl:if test="//@dt:type='uuid'">
  <xsd:simpleType name="uuid">
    <xsd:restriction base="xsd:string">
      <xsd:pattern value="[0-9A-Fa-f]{{8}}\-?[0-9A-Fa-f]{{4}}\-?[0-9A-Fa-f]{{4}}\-?[0-9A-Fa-f]{{4}}\-?[0-9A-Fa-f]{{12}}"/>
    </xsd:restriction>
  </xsd:simpleType>
  </xsl:if>

</xsd:schema>
</xsl:template>

<!-- Library of named templates -->

<xsl:template name="element-content">
  <xsl:apply-templates select="xdr:element | xdr:group | comment()"/>
</xsl:template>

<xsl:template name="attribute-content">
  <xsl:apply-templates select="node()[not(self::xdr:element or self::xdr:group or self::xdr:description or self::xdr:datatype)]"/>
</xsl:template>

<xsl:template name="annotations">
  <xsl:param name="documentation-extras"/>
  <xsl:param name="appinfo-extras"/>
  <xsl:if test="$documentation-extras or xdr:description or *[not(self::xdr:*)]">
    <xsd:annotation>
      <xsl:for-each select="xdr:description">
        <xsd:documentation><xsl:copy-of select="node()"/></xsd:documentation>
      </xsl:for-each>
      <xsl:if test="$documentation-extras">
        <xsd:documentation><xsl:copy-of select="$documentation-extras"/></xsd:documentation>
      </xsl:if>
      <xsl:for-each select="*[not(self::xdr:*)]">
        <xsd:appinfo><xsl:copy-of select="."/></xsd:appinfo>
      </xsl:for-each>
      <xsl:if test="$appinfo-extras">
        <xsd:appinfo><xsl:copy-of select="$appinfo-extras"/></xsd:appinfo>
      </xsl:if>
    </xsd:annotation>
  </xsl:if>    
</xsl:template>

<xsl:template name="any">
  <xsl:if test="not(@model='closed')">
    <xsd:choice minOccurs="0" maxOccurs="unbounded">
      <xsd:any namespace="##other" processContents="lax"/>
      <xsd:any namespace="##targetNamespace"/>
    </xsd:choice>
  </xsl:if>
</xsl:template>

<xsl:template name="any-attribute">
  <xsl:if test="not(@model='closed')">
    <xsd:anyAttribute namespace="##other" processContents="lax"/>
<!-- BUGBUG, can't express this in XSD, as only one anyAttribute is allowed.
    <xsd:anyAttribute namespace="##targetNamespace"/>
-->
  </xsl:if>
</xsl:template>

<xsl:template name="simpleType-with-length">
  <!-- expects the context to be an xdr:ElementType -->
  <xsl:param name="name"/>
  <xsl:variable name="e" select="(. | xdr:datatype)[last()]"/>
  <xsd:simpleType>
    <xsl:if test="$name">
      <xsl:attribute name="name"><xsl:value-of select="$name"/></xsl:attribute>
    </xsl:if>
    <xsd:restriction>
      <xsl:attribute name="base">
        <xsl:choose>
          <xsl:when test="$e/@dt:type">
            <xsl:apply-templates select="$e/@dt:type" mode="convert-datatype"/>
          </xsl:when>
          <xsl:otherwise>xsd:string</xsl:otherwise>
        </xsl:choose>
      </xsl:attribute>
      <xsl:choose>
        <!-- A little optimization - generate "length" when min and max are equal -->
        <xsl:when test="$e/@dt:minLength = $e/@dt:maxLength">
          <xsd:length value="{$e/@dt:minLength}"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:apply-templates select="$e/@dt:minLength | $e/@dt:maxLength"/>
        </xsl:otherwise>
      </xsl:choose>
    </xsd:restriction>
  </xsd:simpleType>
</xsl:template>  <!-- Testcase: "test-attributes.xml" -->

<xsl:template name="simpleType-with-enum">
  <!-- expects the context to be an xdr:ElementType -->
  <xsl:param name="name"/>
  <xsl:variable name="e" select="(. | xdr:datatype)[last()]"/>
  <xsd:simpleType>
    <xsl:if test="$name">
      <xsl:attribute name="name"><xsl:value-of select="$name"/></xsl:attribute>
    </xsl:if>
    <xsd:restriction base="xsd:NMTOKEN">
      <xsl:call-template name="construct-enum">
        <xsl:with-param name="values" select="string($e/@dt:values)"/>
      </xsl:call-template>
    </xsd:restriction>
  </xsd:simpleType>
</xsl:template>

<!-- Recursive template to split dt:values tokens and return an
     <xsd:enumeration> element for each one. -->
<xsl:template name="construct-enum">
  <xsl:param name="values"/>
  <xsl:variable name="first-token" select="substring-before(concat($values,' '), ' ')"/>
  <xsl:variable name="remaining-tokens" select="substring-after($values, ' ')"/>
  <xsl:if test="$first-token">
    <xsd:enumeration value="{$first-token}"/>
  </xsl:if>
  <xsl:if test="$remaining-tokens">
    <xsl:call-template name="construct-enum">
      <xsl:with-param name="values" select="$remaining-tokens"/>
    </xsl:call-template>
  </xsl:if>
</xsl:template>


<!-- element content -->

<xsl:template match="xdr:ElementType">
  <xsl:variable name="e" select="(. | xdr:datatype)[last()]"/>
  
  <xsl:comment> [XDR-XSD] "<xsl:value-of select="@name"/>" element  </xsl:comment>
  
  <!-- In the case of enumerated or datatype allowing a min or max 
       length, and attributes (or an open content model which would
       allow attributes) we need to do a two-step process.  First, 
       generate a simple type with the desired restrictions (enumerated
       values or min and max length).  Later on we can declare a complex
       type derived from this simple type to accomodate the attributes. -->
  <xsl:if test="(@content='textOnly' or $e/dt:type) and (xdr:attribute or not(@model='closed'))">
    <xsl:if test="$e/@dt:type[.='enumeration']">
      <xsl:call-template name="simpleType-with-enum">
        <xsl:with-param name="name" select="concat('generated-type-', generate-id())"/>
      </xsl:call-template>
    </xsl:if>
    <xsl:if test="$e/@dt:maxLength or $e/@dt:minLength">
      <xsl:call-template name="simpleType-with-length">
        <xsl:with-param name="name" select="concat('generated-type-', generate-id())"/>
      </xsl:call-template>
    </xsl:if>
  </xsl:if>
  
  <xsd:element name="{@name}">

    <!-- See if the conditions exist that allows us to define the type
         of the element via a type attribute.  If not, we'll use an
         anonymous type, but that can only happen after we've processed
         the annotaitons.  -->
    <xsl:if test="(@content='textOnly' or $e/@dt:type) and not(xdr:attribute or not(@model='closed'))">
      <xsl:if test="not($e/@dt:type)">
        <xsl:attribute name="type">xsd:string</xsl:attribute>
      </xsl:if>
      <xsl:if test="not($e/@dt:maxLength or $e/@dt:minLength or $e/@dt:type='enumeration')">
        <xsl:apply-templates select="$e/@dt:type"/>
      </xsl:if>
    </xsl:if>  <!-- Testcase: "test-textOnly-content.xml" -->
    <!-- done creating attributes -->
    
    <xsl:call-template name="annotations"/>

    <xsl:choose>
      
      <!-- If an explicit datatype is specified, then the content
           attribute is ignored, and treated as textOnly.  -->
      <xsl:when test="@content='textOnly' or $e/@dt:type">
        <!-- Note that model="open" does not allow elements to be added,
             only attributes. -->
        <xsl:choose>
          <xsl:when test="xdr:attribute or not(@model='closed')">
            <!-- content model allows attributes -->
            <xsd:complexType>
              <xsd:simpleContent>
                <xsd:extension>
                  <xsl:attribute name="base">
                    <xsl:choose>
                      <xsl:when test="$e/@dt:type='enumeration' or $e/@dt:maxLength or $e/@dt:minLength">
                        <!-- Refer to a type generated previously. -->
                        <xsl:text>generated-type-</xsl:text>
                        <xsl:value-of select="generate-id()"/>
                      </xsl:when>
                      <xsl:when test="$e/@dt:type">
                        <xsl:apply-templates select="$e/@dt:type" mode="convert-datatype"/>
                      </xsl:when>
                      <xsl:otherwise>xsd:string</xsl:otherwise>
                    </xsl:choose>
                  </xsl:attribute>
                  <xsl:call-template name="attribute-content"/>
                  <xsl:call-template name="any-attribute"/>
                </xsd:extension>
              </xsd:simpleContent>
            </xsd:complexType>
          </xsl:when>
      
          <xsl:when test="$e/@dt:maxLength or $e/@dt:minLength">
            <xsl:call-template name="simpleType-with-length"/>
          </xsl:when>
          
          <xsl:otherwise>
            <xsl:apply-templates select="@dt:type[.='enumeration'] | xdr:datatype[@dt:type='enumeration']"/>
          </xsl:otherwise>
          
          <!-- General case (simple type, no attributes) is handled
               by generating a type attribute (see beginning of this
               template) instead of simpleType elements.  -->
        </xsl:choose>
      </xsl:when>  <!-- Testcase: "test-textOnly-content.xml" -->
      
      <!-- empty content -->
      <xsl:when test="@content='empty'">
        <xsd:complexType>
          <xsd:complexContent>
            <xsd:restriction base="xsd:anyType">
              <xsl:call-template name="any"/>
              <xsl:call-template name="attribute-content"/>
              <xsl:call-template name="any-attribute"/>
            </xsd:restriction>
          </xsd:complexContent>
        </xsd:complexType>
      </xsl:when>  <!-- Testcase: "test-empty-content.xml" -->
      
      <!--  If no content attribute is specified, but the order 
            attribute is, XDR acts as if @content defaulted to 
            eltOnly. -->
      <xsl:when test="@content='eltOnly' or @order='one' or @order='seq'">
        <xsd:complexType>
          <xsl:choose>

            <xsl:when test="@order='one'">
              <xsl:choose>
                <xsl:when test="not(@model='closed')">
                  <!-- This is an unlikely case - an open model conflicts 
                       with the order='one' restriction, which is likely to 
                       produce results unexpected by the user.  XDR 
                       documentation warns against this combination, but 
                       allows it.  Thus the behavior is duplicated here. -->
                  <xsd:sequence>
                    <xsd:choice>
                      <xsl:call-template name="element-content"/>
                    </xsd:choice>
                    <xsl:call-template name="any"/>
                  </xsd:sequence>
                  <xsl:call-template name="attribute-content"/>
                  <xsl:call-template name="any-attribute"/>
                </xsl:when>
                <xsl:otherwise>
                  <!-- model is closed -->
                  <xsd:choice>
                    <xsl:call-template name="element-content"/>
                  </xsd:choice>
                  <xsl:call-template name="attribute-content"/>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:when>
            
            <xsl:when test="@order='seq'">
              <xsd:sequence>
                <xsl:call-template name="element-content"/>
                <xsl:call-template name="any"/>
              </xsd:sequence>
              <xsl:call-template name="attribute-content"/>
              <xsl:call-template name="any-attribute"/>
            </xsl:when>

            <xsl:otherwise>  <!-- @order='many' -->
              <xsd:choice minOccurs="0" maxOccurs="unbounded">
                <xsl:call-template name="element-content"/>
                <xsl:call-template name="any"/>
              </xsd:choice>
              <xsl:call-template name="attribute-content"/>
              <xsl:call-template name="any-attribute"/>
            </xsl:otherwise>
          </xsl:choose>
        </xsd:complexType>
      </xsl:when>  <!-- Testcase: "test-eltOnly-content.xml" -->
      
      <!-- content='mixed' or is not specified, in which case it
           defaults to 'mixed' if the order attribute has the value
           'many' or is not specified. -->
      <xsl:otherwise>
        <xsl:if test="@order='one' or @order='seq'">
          <xsl:comment>
            [XDR-XSD] ****Warning!****
            Original schema illegally combined content="mixed" and @order="<xsl:value-of select="@order"/>".
            Treating this as order='many' instead.
          </xsl:comment>
        </xsl:if>
        <xsd:complexType mixed="true">
          <xsd:choice maxOccurs="unbounded">
            <xsl:call-template name="element-content"/>
            <xsl:call-template name="any"/>
          </xsd:choice>
          <xsl:call-template name="attribute-content"/>
          <xsl:call-template name="any-attribute"/>
        </xsd:complexType>
      </xsl:otherwise>  <!-- Testcase: "test-mixed-content.xml" -->
      
    </xsl:choose>
  </xsd:element>
  
</xsl:template>

<xsl:template match="xdr:element">
  <xsd:element ref="{@type}">
    <xsl:apply-templates select="@minOccurs | @maxOccurs"/>
    <xsl:call-template name="annotations"/>
  </xsd:element>
</xsl:template>

<xsl:template match="xdr:group">
  <!-- Trace the inheritance to discover what "order" value applies -->
  <xsl:variable name="inherited-order">
    <xsl:choose>
      <xsl:when test="ancestor-or-self::xdr:group[@order]">
        <xsl:value-of select="ancestor-or-self::xdr:group[@order][1]/@order"/>
      </xsl:when>
      <xsl:when test="ancestor::xdr:ElementType[@order]">
        <xsl:value-of select="ancestor::xdr:ElementType[@order]/@order"/>
      </xsl:when>
      <xsl:when test="ancestor::xdr:ElementType/@content='eltOnly'">seq</xsl:when>
      <xsl:otherwise>many</xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <!-- Translate "order" into the name of a group -->
  <xsl:variable name="group-name">
    <xsl:choose>
      <xsl:when test="$inherited-order='one'">choice</xsl:when>
      <xsl:when test="$inherited-order='seq'">sequence</xsl:when>
      <xsl:when test="$inherited-order='many'">choice</xsl:when>
    </xsl:choose>
  </xsl:variable>

  <xsl:element name="xsd:{$group-name}">
    <xsl:choose>
      <xsl:when test="$inherited-order='many'">
        <xsl:attribute name="minOccurs">0</xsl:attribute>
        <xsl:attribute name="maxOccurs">unbounded</xsl:attribute>
      </xsl:when>
      <xsl:otherwise>
        <xsl:apply-templates select="@minOccurs | @maxOccurs"/>
      </xsl:otherwise>
    </xsl:choose>
    <xsl:call-template name="annotations"/>
    <xsl:call-template name="element-content"/>
  </xsl:element>
</xsl:template>

<!-- attribute content -->

<!-- Generate global attributes, even if these aren't referenced anywhere. -->
<xsl:template match="xdr:Schema/xdr:AttributeType">
  <xsd:attribute name="{@name}">
    <xsl:apply-templates select="@required"/>
    <xsl:apply-templates select="@default"/>

    <!-- if required="yes" and default is specified, this is equivalent to 
         a fixed attribute.  -->
    <xsl:if test="@required='yes' and @default">
      <!-- <xsl:attribute name="use">fixed</xsl:attribute> 2000 schema style -->
      <xsl:attribute name="use">required</xsl:attribute>
    </xsl:if>
        
    <!-- enumerations and types with a specified minLength or maxLength cannot be
         expressed through XSD attributes, so don't handle them yet.  -->
    <xsl:if test="not(@dt:maxLength or xdr:datatype/@dt:maxLength or @dt:minLength or xdr:datatype/@dt:minLength or @dt:type[.='enumeration'] or xdr:datatype[@dt:type='enumeration'])">
      <xsl:apply-templates select="@dt:type | xdr:datatype"/>
    </xsl:if>
    <!-- done generating attributes -->

    <!-- description and annotations -->
    <xsl:apply-templates select="comment()"/>
    <xsl:call-template name="annotations"/>
    
    <!-- generate an anonymous type with the enumerated values -->
    <xsl:apply-templates select="@dt:type[.='enumeration'] | xdr:datatype[@dt:type='enumeration']"/>

    <!-- generate an anonymous type with the minLength or maxLength restrictions -->
    <xsl:if test="@dt:minLength or xdr:datatype/@dt:minLength or @dt:maxLength or xdr:datatype/@dt:maxLength">
      <xsl:call-template name="simpleType-with-length"/>
    </xsl:if>
  </xsd:attribute>
</xsl:template>

<!-- Local attribute definitions are preserved as local content types -->
<xsl:template match="xdr:AttributeType"/>

<xsl:template match="xdr:attribute">
  <!-- Find the applicable (closest) AttributeType definition. -->
  <xsl:variable name="definition" select="ancestor::*[xdr:AttributeType/@name = current()/@type][1]/xdr:AttributeType[@name = current()/@type]"/>

  <xsl:choose>
    <!--
      Note that XDR allows attribute references to be further refined (default 
      values) more heavily than XSD, so when default values appear on the
      attribute declaration, the information must be copied from the global
      attribute.
    -->
    <xsl:when test="not(@default) and $definition/parent::xdr:Schema">
      <xsd:attribute ref="{@type}">
        <xsl:apply-templates select="@required"/>
        <xsl:call-template name="annotations"/>
      </xsd:attribute>
    </xsl:when>

    <xsl:otherwise>
      <xsd:attribute name="{@type}">

        <!-- required attribute on the attribute declaration override that on the definition -->
        <xsl:choose>
          <xsl:when test="@required"><xsl:apply-templates select="@required"/></xsl:when>
          <xsl:otherwise><xsl:apply-templates select="$definition/@required"/></xsl:otherwise>
        </xsl:choose>

        <!-- default attribute on the attribute declaration override that on the definition -->
        <xsl:choose>
          <xsl:when test="@default"><xsl:apply-templates select="@default"/></xsl:when>
          <xsl:otherwise><xsl:apply-templates select="$definition/@default"/></xsl:otherwise>
        </xsl:choose>

        <!-- if required="yes" and default is specified, this is equivalent to 
             a fixed attribute.  -->
        <xsl:if test="($definition/@required='yes' or @required='yes') and ($definition/@default or @default)">
          <!-- <xsl:attribute name="use">fixed</xsl:attribute> 2000 schema style-->
          <xsl:attribute name="use">required</xsl:attribute>
        </xsl:if>
        
        <!-- enumerations and types with a specified minLength or maxLength cannot be
             expressed through XSD attributes, so don't handle them yet.  -->
        <xsl:if test="not($definition[@dt:maxLength or xdr:datatype/@dt:maxLength or @dt:minLength or xdr:datatype/@dt:minLength or @dt:type[.='enumeration'] or xdr:datatype[@dt:type='enumeration']])">
          <xsl:apply-templates select="$definition/@dt:type | $definition/xdr:datatype"/>
        </xsl:if>
        <!-- done generating attributes -->

        <xsl:apply-templates select="$definition/comment()"/>
        <!-- description and annotations -->
        <xsl:variable name="local-annotations" select="*[not(self::xdr:*)]"/>
        <xsl:for-each select="$definition">
          <xsl:call-template name="annotations">
            <xsl:with-param name="appinfo-extras" select="$local-annotations"/>
          </xsl:call-template>
        </xsl:for-each>  <!-- Testcase: "test-annotations.xml" -->
          
        <!-- generate an anonymous type with the enumerated values -->
        <xsl:apply-templates select="$definition/@dt:type[.='enumeration'] | $definition/xdr:datatype[@dt:type='enumeration']"/>

        <!-- generate an anonymous type with the minLength or maxLength restrictions -->
        <xsl:for-each select="$definition[@dt:minLength or @dt:maxLength or xdr:datatype[@dt:minLength or @dt:maxLength]]">
          <xsl:call-template name="simpleType-with-length"/>
       </xsl:for-each>
      </xsd:attribute>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>  <!-- Testcase: "test-attributes.xml" -->

<!-- other constructs -->

<!-- descriptions are handled elsewhere, ignore any 
     that fell through the cracks -->
<xsl:template match="xdr:description"/>

<!--  attribute conversion templates -->

<xsl:template match="@*"/>

<xsl:template match="@dt:minLength">
  <xsd:minLength value="{.}"/>
</xsl:template>

<xsl:template match="@dt:maxLength">
  <xsd:maxLength value="{.}"/>
</xsl:template>

<xsl:template match="@minOccurs | @maxOccurs">
  <xsl:copy/>
</xsl:template>

<xsl:template match="@maxOccurs[.='*']">
  <xsl:attribute name="maxOccurs">unbounded</xsl:attribute>
</xsl:template>

<xsl:template match="@default">
  <xsl:attribute name="use">default</xsl:attribute>
  <!-- <xsl:attribute name="value"><xsl:value-of select="."/></xsl:attribute> 2000 schema style -->
  <xsl:attribute name="fixed"><xsl:value-of select="."/></xsl:attribute>
</xsl:template>

<xsl:template match="@required[.='yes']">
  <xsl:attribute name="use">required</xsl:attribute>
</xsl:template>

<!--  Data type attribute/element conversions -->

<xsl:template match="@dt:type">
  <xsl:attribute name="type">
    <xsl:apply-templates select="." mode="convert-datatype"/>
  </xsl:attribute>
</xsl:template>

<xsl:template match="xdr:datatype">
  <xsl:apply-templates select="@dt:type"/>
</xsl:template>

<xsl:template match="@dt:type[.='enumeration']">
  <xsl:for-each select="..">
    <xsl:call-template name="simpleType-with-enum"/>
  </xsl:for-each>
</xsl:template>  <!-- Testcase: "test-attributes.xml" -->

<!--
  convert-datatype mode just returns the converted datatype value, instead
  of an attribute with the converted value.  These are special cases needed
  for dt:minLength, dt:maxLength conversions.
-->

<!-- no conversion for enumerations, which must be handled separately -->
<xsl:template mode="convert-datatype" match="@dt:type[.='enumeration']">xsd:enumeration</xsl:template>

<xsl:template mode="convert-datatype" match="@dt:type"><xsl:value-of select="."/></xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='id']">xsd:ID</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='idref']">xsd:IDREF</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='idrefs']">xsd:IDREFS</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='entity']">xsd:ENTITY</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='entities']">xsd:ENTITIES</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='nmtoken']">xsd:NMTOKEN</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='nmtokens']">xsd:NMTOKENS</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='notation']">xsd:NOTATION</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='int']">xsd:integer</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='i1']">xsd:byte</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='i2']">xsd:short</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='i4']">xsd:int</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='i8']">xsd:long</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='ui1']">xsd:unsignedByte</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='ui2']">xsd:unsignedShort</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='ui4']">xsd:unsignedInt</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='ui8']">xsd:unsignedLong</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='r8']">xsd:double</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='boolean']">ISOBoolean</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='date.tz']">xsd:date</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='dateTime.tz']">xsd:timeInstant</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='time.tz']">xsd:time</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='float']">xsd:float</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='uri']">xsd:uriReference</xsl:template>
<xsl:template mode="convert-datatype" match="@dt:type[.='string']">xsd:string</xsl:template>

</xsl:stylesheet>